# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
""""model file"""
import logging
import tensorflow as tf
from tensorflow.keras import Model, layers


logger = logging.getLogger()


class FernModel(object):
    model: Model

    def __init__(self, output_shape, max_seq_len, library_len, initializer='he_normal'):
        """
        model builder

        Parameters
        ----------
        output_shape : dict[str, int], list[int], tuple[int]
            output shape without batch size
        max_seq_len : int
            the max input sequence length
        library_len : int
            the world library length
        initializer : str
            global initializer
        """
        self.output_shape = output_shape
        self.max_seq_len = max_seq_len
        self.library_len = library_len
        self.initializer = initializer
        self.name = self.__class__.__name__

        self.model = self.build()
        self.print_summary()

        self.compile = self.model.compile
        self.fit = self.model.fit

    def print_summary(self):
        """
        print summary of model
        """
        summary = []
        self.model.summary(print_fn=summary.append)
        summary = '\n'.join(summary)
        logger.info(f"\n{summary}")

    def build(self):
        """
        build model

        Returns
        -------
        Model
            built model
        """
        raise NotImplementedError

    def save(self, path):
        """
        save model

        Parameters
        ----------
        path : str, pathlib.Path
            The model file path
        """
        self.model.save(path)

    def load(self, path):
        """
        load model

        Parameters
        ----------
        path : str, pathlib.Path
            The model file path
        """
        self.model.load_weights(path)

    @property
    def predict(self):
        return self.model.predict

    def __call__(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    @property
    def trainable_variables(self):
        return self.model.trainable_variables


class TextCNN(FernModel):
    """
    References
    ----------
    Optimization of model Convolutional Neural Networks for Sentence Classification
     (https://arxiv.org/pdf/1408.5882.pdf)
    """
    def build(self):
        inp = layers.Input(shape=(self.max_seq_len,))
        x = layers.Embedding(self.library_len, 256, embeddings_initializer=self.initializer)(inp)
        x = layers.Conv1D(256, 5, padding='same', kernel_initializer=self.initializer)(x)
        x = layers.BatchNormalization()(x)
        x = layers.Activation('relu')(x)

        x = layers.GlobalMaxPool1D()(x)

        x = layers.Dense(128, kernel_initializer=self.initializer)(x)
        x = layers.BatchNormalization()(x)
        x = layers.Activation('relu')(x)

        ys = []
        for key in self.output_shape:
            y = layers.Dense(self.output_shape[key],
                             kernel_initializer=self.initializer,
                             activation='softmax',
                             name=key)(x)
            ys.append(y)
        model = Model(inputs=inp, outputs=ys, name=self.name)    # if len(ys) == 1, than ys = ys[0]
        return model


class Word2Vec(tf.Module):
    """
    根据论文在TF2.0上复现word2vec模型, 这里主要复现Skip-Gram模型, 该模型得到的结论：

    1. 当前词和附近词相似（余弦相似度）
    2. 根据这个训练方法得到的词向量直接相互加减可以表达部分语义


    References:
        1. [TF2.0 word2vec](https://www.tensorflow.org/tutorials/text/word2vec#subclassed_word2vec_model)

    """
    def __init__(self, vocab_size: int = 10000, embedding_dim: int = 128, num_ns: int = 4):
        super().__init__(name='word2vec')
        self.vocab_size = vocab_size
        self.num_ns = num_ns
        self.target_embedding = layers.Embedding(
            vocab_size,
            embedding_dim,
            input_length=1,
            name="w2v_embedding")
        self.context_weight = tf.Variable(tf.random.normal((vocab_size, embedding_dim)), name='context_weight')
        self.context_bias = tf.Variable(tf.random.normal((vocab_size, )), name='context_bias')  # embed的需要为每个词添加bias

    @tf.function
    def __call__(self, target):
        """
        计算词列表的词向量

        Args:
            target: 一维的词id列表, shape=(batch,)

        Returns:
            返回shape=(batch, embed)的词向量
        """
        target = self.target_embedding(target)  # (batch, embed)
        return target

    @tf.function
    def loss(self, target, context, training=None):
        """训练时使用NCE loss，否则使用sigmoid cross entropy loss

        Args:
            target: 经过embedding后的词向量
            context: 上下文词的编号，起始可以理解为label
            training: 如果训练，返回nce loss，否则返回sigmoid loss

        Returns:
            返回shape=()的loss

        """
        # target, context  # (batch, embed)  (batch,)
        if training:
            loss = tf.reduce_mean(tf.nn.nce_loss(weights=self.context_weight, biases=self.context_bias, labels=context,
                                                 inputs=target, num_sampled=self.num_ns, num_classes=self.vocab_size))
        else:
            logits = tf.matmul(target, tf.transpose(self.context_weight))
            logits = tf.nn.bias_add(logits, self.context_bias)
            labels_one_hot = tf.one_hot(context, self.vocab_size)
            labels_one_hot = tf.squeeze(labels_one_hot, axis=1)
            loss = tf.nn.sigmoid_cross_entropy_with_logits(
                labels=labels_one_hot,
                logits=logits)
            loss = tf.reduce_mean(loss)
        return loss

    def save(self, model_dir):
        """保存词向量到本地"""
        tf.saved_model.save(self, str(model_dir))

    @classmethod
    def load(cls, model_dir):
        """从配置文件中加载重新加载模型"""
        model = tf.saved_model.load(model_dir)
        return model



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""
tools for files in hdfs and local file system
"""
from typing import *
import os
import re
import pathlib
import subprocess

from fern.logging import Logging


def is_hdfs_path(path: str):
    """检查path是否是合法的HDFS path, 合法的hdfs path格式: hdfs://a/..."""
    paths = re.split(r'(/+)', path)
    if ('hdfs' not in path or '..' in path or './' in path) or (len(paths) < 4):
        ret = False
    else:
        ret = True
    return ret


class HDFSPurePath(object):
    """只做HDFS路径操作"""
    def __init__(self, path: str):
        path = re.sub(r'/+$', '', path)
        self._path = path
        self._root = self._get_root(path)
        self._relative_path = self._get_relative_path(path)
        self.logger = Logging()

    def __new__(cls, path, *args, **kwargs):
        if isinstance(path, HDFSPurePath):
            self = path
        else:
            self = object.__new__(cls)
            self = cls.__init__(self, path)
        return self

    @classmethod
    def _make_path(cls, root: str = None, relative_path: str = None, abspath: str = None):
        """根据参数创建新的path

        Args:
            root: format 'hdfs://xxx'
            relative_path: format 'a/b' or '.'
            abspath: format 'hdfs://xx/a/b'
        
        Returns:
            PurePath
        """
        if abspath is None:
            if relative_path == '.':
                abspath = root
            else:
                abspath = root + '/' + relative_path
        return cls(abspath)

    @staticmethod
    def _get_root(path) -> str:
        """读取path的root目录: hdfs://a/b/c -> hdfs://a"""
        ret = re.findall(r'^hdfs://[^/]+', path)
        assert len(ret) == 1, f'Unable to find the root path from: {path}'
        ret = ret[0]
        return ret

    def _get_relative_path(self, path) -> str:
        """读取hdfs path除了根目录外的相对路径, 用于所有的路径操作

        Returns:
            仅返回如下格式: ['a/b', '.']
        """
        root = self._get_root(path)
        path = re.sub(rf'^{root}', r'.', path)
        path = pathlib.PosixPath(path)
        path = str(path)
        return path

    @property
    def root(self):
        return self._root

    @property
    def parent(self):
        """获取当前路径的父路径"""
        relative_path = str(pathlib.PosixPath(self._relative_path).parent)
        return self._make_path(root=self._root, relative_path=relative_path)
    
    @property
    def name(self) -> str:
        """a/b.c -> b.c"""
        return pathlib.PosixPath(self._relative_path).name

    @property
    def stem(self) -> str:
        """a/b.c -> b"""
        return pathlib.PosixPath(self._relative_path).stem
    
    @property
    def suffix(self) -> str:
        """a/b.c -> .c"""
        return pathlib.PosixPath(self._relative_path).suffix

    def __repr__(self):
        """打印路径文本: for str function and print function"""
        relative_path = self._relative_path
        ret = [self._root]
        if relative_path != '.':
            ret.append(relative_path)
        ret = '/'.join(ret)
        return ret
    
    def __truediv__(self, other):
        """hdfs://a / b => hdfs://a/b"""
        relative_path = str(pathlib.PosixPath(self._relative_path) / str(other))
        return self._make_path(root=self._root, relative_path=relative_path)
    
    def __eq__(self, other):
        """a == b"""
        assert isinstance(other, HDFSPurePath)
        return str(self) == str(other)


class HDFSPath(HDFSPurePath):
    """实现HDFS文件常用查询"""
    def exists(self) -> bool:
        ret = False
        with subprocess.Popen(['hdfs', 'dfs', '-ls', str(self)], stderr=subprocess.PIPE).stderr as oup:
            for line in oup:
                line = line.decode('utf-8')
                if 'No such file or directory' in line:
                    ret = True
                    break
        return ret

    def is_dir(self) -> bool:
        return self.exists() and not self.is_file()

    def is_file(self) -> bool:
        if not self.exists():
            return False
        ret = False
        if self.parent != self:
            # not root dir
            name = self.name
            with subprocess.Popen(['hdfs', 'dfs', '-ls', str(self.parent)], stdout=subprocess.PIPE).stdout as oup:
                for line in oup:
                    # line: b'-rw-r--r-- 3 hive supergroup 0 2022-05-18 06:35
                    # hdfs://ana-hdfs/tmp/ai_tech/max/sql_test/_SUCCESS\n'
                    line = line.decode('utf-8')
                    if 'hdfs' in line and name in line:
                        # current file or folder
                        path_type = line.strip().split()[0]
                        ret = path_type[0] != 'd'  # drw-r--r--
                        break
        return ret

    def glob(self, pattern: str) -> Generator:
        """递归当前目录以及其子目录， 当前目录必须是文件夹

        Args:
            pattern: 支持的pattern模式， 和pathlib中的匹配规则一致
        """
        with subprocess.Popen(['hdfs', 'dfs', '-ls', f'{self._path}/{pattern}'], stdout=subprocess.PIPE).stdout as oup:
            for line in oup:
                line = line.decode('utf-8')
                if 'hdfs' in line:
                    path = line.strip().split()[-1]
                    path = self._make_path(abspath=path)
                    yield path


class Path(HDFSPath):
    def __new__(cls, path: str):
        if is_hdfs_path(path):
            self = object.__new__(cls)
            cls.__init__(self, path)
        else:
            self = pathlib.Path(path)
        return self



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""存放部分模型现成的数据流水线功能"""
from typing import *
from pathlib import Path
from collections import defaultdict

import pandas as pd
import tensorflow as tf

from fern.data import read_words, train_data_split, str2word


class BaseDataPipeline(object):
    batch_size: Optional[int] = None

    def get_dataset(self, training: bool) -> tf.data.Dataset:
        raise NotImplementedError

    @property
    def train_ds(self):
        return self.get_dataset(training=True)

    @property
    def val_ds(self):
        return self.get_dataset(training=False)


class Word2VecDataPipeline(BaseDataPipeline):
    """训练word2vec向量(skip-gram)，做了如下工作：

    1. 通过输入的数据创建数据流水线
    2. 实现停用词过滤
    3. 实现分词
    4. 实现大小写配置
    5. 创建词库, 忽略词频<2的单词

    """
    word2id: Optional[Dict] = None
    unk_token = '[UNK]'

    def __init__(self,
                 data: List[str],
                 batch_size: int,
                 stop_words: str = None,
                 zh_segmentation: bool = True,
                 lower: bool = True,
                 win_size: int = 2,
                 random_seed: Optional[int] = None,
                 ):
        """
        Args:
            data: 原始数据列表
            batch_size: 批大小
            stop_words: 停用词路径
            zh_segmentation: 中文是否使用分词工具分词
            lower: 是否忽略大小写
            win_size: 生成(target, context)的窗口大小
            random_seed: 随机种子，用于复现
        """
        self._data = pd.DataFrame({'source': data})
        if lower:
            self._data['source'] = self._data['source'].map(lambda x: x.lower())

        if stop_words is None:
            stop_words = Path(__file__).parent / 'config/stop_words.txt'
        self.stop_words = read_words(stop_words)
        self.word2id, self.id2word = self._build_vocab(zh_segmentation)
        self.vocab_size = len(self.word2id)

        # 分词，词转id，分数据集
        self._data['tokens'] = self._data['source'].map(lambda x: self.tokenize(x, zh_segmentation))
        self._data['ids'] = self._data['source'].map(self.token2id)
        self.train_data, self.val_data = train_data_split(self._data, test_size=0.2, random_state=random_seed)

        # 数据生成其他配置
        self.win_size = win_size
        self.batch_size = batch_size

    def _build_vocab(self, zh_segmentation: bool):
        """制作的词库按照词频从大到小排列, UNK放在最后一个字符"""
        def map_func(sentence: str):
            words = str2word(sentence, zh_segmentation)
            for _word in words:
                if _word in self.stop_words:
                    continue
                vocab[_word] += 1

        vocab = defaultdict(int)
        self._data['source'].map(map_func)
        vocab = {word: value for word, value in vocab.items() if value > 1}
        vocab = sorted(vocab, key=lambda key: vocab[key], reverse=True)
        vocab = list(vocab) + [self.unk_token]
        word2id = {word: idx for idx, word in enumerate(vocab)}
        id2word = {idx: word for idx, word in enumerate(vocab)}
        return word2id, id2word

    def tokenize(self, sentence: str, zh_segmentation: bool) -> List[str]:
        """把句子转化为token数据，如果不在词库中，那么会替换为UNK"""
        assert self.word2id is not None
        words = str2word(sentence, zh_segmentation)
        for i in range(len(words)):
            if words[i] not in self.word2id:
                words[i] = self.unk_token
        return words

    def token2id(self, words: List[str]) -> List[int]:
        """把词列表转化为token列表"""
        ids = [self.word2id.get(word, self.word2id[self.unk_token]) for word in words]
        return ids

    def data_generator(self, data: pd.DataFrame):
        """生成(target_id, context_id)"""
        data = data.sample(frac=1)
        for index, row in data.iterrows():
            # 每行一个句子，根据win_size从句子中生成（target, context）
            ids = row['ids']
            for i in range(len(ids)):
                index_min = max(i - self.win_size, 0)
                index_max = min(i + self.win_size, len(ids) - 1)
                for ii in range(index_min, index_max+1):
                    if ii == i:
                        continue
                    yield ids[i], [ids[ii]]

    def get_dataset(self, training: bool):
        """
        把生成器转化为tensorflow数据集

        Args:
            training: True返回训练数据集，如果是False返回测试数据集

        Returns:
            tensorflow数据集
        """
        if training:
            data = self.train_data
        else:
            data = self.val_data
        target_id = tf.TensorSpec(shape=(), dtype=tf.int64)
        context_id = tf.TensorSpec(shape=(1,), dtype=tf.int64)
        output_signature = (target_id, context_id)
        dataset = tf.data.Dataset.from_generator(
            generator=lambda: self.data_generator(data),
            output_signature=output_signature)
        dataset = dataset.cache().shuffle(self.batch_size * 10000).batch(self.batch_size).prefetch(tf.data.AUTOTUNE)
        return dataset



# Fern
# 
# @Author: Jason, Lin
# @Email : Jason.M.Lin@outlook.com
# @Time  : 2022/2/9 9:15 上午
#
# =============================================================================
"""特定模型直接训练调度代码，同时也是为了解耦model&train代码的嵌套import。几个模块之间的关系如下

Estimator:
    Data Pipeline
    Model
    Trainer
"""
from typing import *
from fern.train import Word2VecTrainer
from fern.models import Word2Vec
from fern.pipeline import Word2VecDataPipeline


def word2vec_estimator(
        data: List[str],
        batch_size: int,
        zh_segmentation: bool,
        model_home: str,
        ckpt_home: str,
        log_home: str,
        epoch: int,
        embedding_dim: int = 128,
        version: int = 1,
        opt: str = 'adam',
        lr: float = 1.0e-3,
        lower: bool = True,
        win_size: int = 2,
        random_seed: int = 42,
        verbose: int = 1) -> Tuple[Word2Vec, List[str], float]:
    """调用对应的训练器训练模型

    Examples:
        >>> # 训练
        >>> import tensorflow as tf
        >>> from fern.estimate import word2vec_estimator
        >>> path_to_file = tf.keras.utils.get_file('shakespeare.txt',
        ...     'https://storage.googleapis.com/download.tensorflow.org/data/shakespeare.txt')
        >>> with open(path_to_file, 'r') as f:
        ...     data = []
        ...     for line in f:
        ...         if line.strip():
        ...             data.append(line.strip())
        >>> model, vocab, loss = word2vec_estimator(
        ...     data=data[:100],
        ...     batch_size=128,
        ...     zh_segmentation=True,
        ...     model_home='/tmp/Fern/model',
        ...     ckpt_home='/tmp/Fern/ckpt',
        ...     log_home='/tmp/Fern/log',
        ...     epoch=200,
        ...     embedding_dim=8,
        ...     version=1,
        ...     opt='adam',
        ...     lr=1.0e-3,
        ...     lower=True,
        ...     win_size=2,
        ...     random_seed=42,
        ...     verbose=1)
        >>> print(loss)
        >>> # 获取词向量
        >>> model.target_embedding.variables[0].numpy()

    Args:
        data: 训练数据，每个元素都是一个句子
        batch_size: 每批次大小
        zh_segmentation: 是否做中文分词, 否则中文按字分割
        model_home: 模型保存根目录
        ckpt_home: 模型检查点目录
        log_home: 日志检查点目录
        epoch: 循环训练多少次
        embedding_dim: word2vec的词向量大小
        version: 模型的版本号
        opt: 优化器
        lr: 学习率
        lower: 是否忽略大小写
        win_size: 生成(target, context)的窗口大小
        random_seed: 随机种子大小
        verbose: 0: 不打印日志；1打印epoch日志；2: 打印batch日志

    Returns:
        (model, vocab, train_loss): 模型，词库，训练损失
    """
    pipeline = Word2VecDataPipeline(data=data, batch_size=batch_size, zh_segmentation=zh_segmentation, lower=lower,
                                    win_size=win_size, random_seed=random_seed)
    train_ds = pipeline.train_ds
    val_ds = pipeline.val_ds

    model = Word2Vec(vocab_size=pipeline.vocab_size, embedding_dim=embedding_dim)

    trainer = Word2VecTrainer(model=model, model_home=model_home, ckpt_home=ckpt_home, log_home=log_home,
                              train_ds=train_ds, val_ds=val_ds, opt=opt, lr=lr, version=version)
    train_loss = trainer.train(epoch=epoch, verbose=verbose)
    return model, list(pipeline.word2id.keys()), train_loss
